// This may look like C code, but it is really -*- C++ -*-
/****************************************************************************
** $Id: fancysockdevice.h 191 2009-02-20 21:19:59Z wglas $
**
** Copyright (C) 2003 EV-i Informationstechnologie GmbH
**
** Author: Wolfgang Glas / EV-i
**
** A base class, which wraps TCP/IP sockets for use with libxml.
** Extensible for using SSL or socks etc. by using many virtual functions.
** 
*****************************************************************************/

#ifndef FANCYSOCK_DEVICE_H
#define FANCYSOCK_DEVICE_H

#ifdef WIN32
#define AF_LOCAL 99
extern "C" {
  struct sockaddr;
  struct sockaddr_in;
  struct sockaddr_un
  {
    short sun_family;
    char  sun_path[108];
  };
  // define this in order to avoid inclusion of winsock2.h
  struct sockaddr_in_local
  {
    short          sin_family;
    unsigned short sin_port;
    unsigned char  sin_data[12];
  };
  typedef int socklen_t;
}
#define POLLIN      0x0001
#define POLLPRI     0x0002
#define POLLOUT     0x0004
#define POLLERR     0x0008
#define POLLHUP     0x0010
#define POLLNVAL    0x0020
#else
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>
#endif
#include <string>
#include <iosfwd>
#include <fancysock.h>

namespace Fancy {
  namespace Sock {

    class FANCYSOCK_API Address
    {
    protected:
      Address() {}

    public:
      virtual ~Address();

      virtual const struct sockaddr *sockAddr() const = 0;
      virtual socklen_t sockLen() const = 0;
      virtual int addrDomain() const = 0;

      virtual Address *dup() const = 0;
      virtual std::string toString() const = 0;
      virtual void setString (const std::string &address) = 0;

      virtual bool isNull() const = 0;

      static Address *create (const struct sockaddr *addr, socklen_t len);
      static Address *createFromString (const std::string &address);
      static Address *resolve (const std::string &name, int def_port=0);
    };

    class FANCYSOCK_API IP4Address : public Address
    {
    public:
      IP4Address ();
      IP4Address (int port, unsigned long addr = 0);
      IP4Address (const struct sockaddr_in *_addr);
      IP4Address (const IP4Address &a);

      ~IP4Address();

      Address *dup() const;
      std::string toString() const;
      void setString (const std::string &address);

      bool isNull() const;

      const struct sockaddr *sockAddr() const;
      socklen_t sockLen() const;
      int addrDomain() const;

    private:
#ifdef WIN32
      struct sockaddr_in_local addr;
#else
      struct sockaddr_in addr;
#endif
    };

    class FANCYSOCK_API LocalAddress : public Address
    {
    public:
      LocalAddress ();
      LocalAddress (const std::string &path);
      LocalAddress (const struct sockaddr_un *_addr);
      LocalAddress (const LocalAddress &a);

      ~LocalAddress();

      Address *dup() const;
      std::string toString() const;
      void setString (const std::string &path);

      bool isNull() const;

      const struct sockaddr *sockAddr() const;
      socklen_t sockLen() const;
      int addrDomain() const;

    private:
      struct sockaddr_un addr;
    };

    class FANCYSOCK_API Device
    {
    public:
      Device();

      virtual ~Device();

      // client stuff.
      virtual void connect (const Address &addr);

      // server stuff.
      void bind (const Address &addr);
      virtual void listen (int backlog);
      virtual Device *accept();

      // common functions.
      int poll (short *events, int timeout);

      virtual void close();
 
      size_t bytesAvailable() const;
 
      virtual int read (void *buf, int count);
      virtual int write (const void *buf, int count);

      const Address *getAddr();

      virtual void statusInfo(std::ostream &os) const;

#ifdef WIN32
      typedef void *sock_handle_t;
#else
      typedef int sock_handle_t;
#endif

      class FANCYSOCK_API AcceptHandle
      {
      public:
        AcceptHandle(Address *a, sock_handle_t fd);
        AcceptHandle(const AcceptHandle &a);

        AcceptHandle &operator=(const AcceptHandle &a);

        ~AcceptHandle();

        const Address *getAddress() const { return addr; }
        sock_handle_t getHandle() const { return fd; }

        Address *takeAddress() const { Address *ret = addr; addr=0; return ret; }
        sock_handle_t takeHandle() const;

      private:
        mutable sock_handle_t fd;
        mutable Address *addr;
      };

    protected:
      AcceptHandle accept_fd();
      Device(const AcceptHandle &a);

      int getSockHandle() const;

      sock_handle_t fd;
      Address *addr;
      bool listening;
    };
  }
}

#endif // ! FANCYSOCK_DEVICE_H
